package com.jclark.xsl.sax;

import com.jclark.xsl.sax.*;
import java.io.*;
import org.xml.sax.*;
import java.util.Hashtable;

public class XhtmlOutputHandler implements OutputDocumentHandler, CommentHandler, RawCharactersHandler {

// Copied from HTMLOutputHandler to handle empty elements

  static private final int EMPTY_CONTENT = 1;

  static private final String emptyElements[] = {
    "area",
    "base",
    "basefont",
    "br",
    "col",
    "frame",
    "hr",
    "img",
    "input",
    "isindex",
    "link",
    "meta",
    "param"
  };

  static private Hashtable elementTypeTable = new Hashtable();

	static {
		for (int i = 0; i < emptyElements.length; i++) {
		elementTypeTable.put(emptyElements[i], new Integer(EMPTY_CONTENT));
		}
	}

  private static int getElementTypeFlags(String name) {
    Integer type
      = (Integer)elementTypeTable.get(name.toLowerCase());
    if (type == null)
      return 0;
    return type.intValue();
  }

// Copied from XMLOutputHandler which can't be extended since may methods are private


  private Writer writer;
  private char maxRepresentableChar = '\uFFFF';
  private boolean keepOpen;
  private boolean inStartTag = false;
  private boolean omitXmlDeclaration = false;
  private String standalone;
  private static final int DEFAULT_BUF_LENGTH = 8*1024;
  private char[] buf = new char[DEFAULT_BUF_LENGTH];
  private int bufUsed = 0;
  private String lineSeparator;
  private byte minimize = MINIMIZE_EMPTY_ELEMENTS;
  private String doctypeSystem;
  private String doctypePublic;
  private boolean outputDoctype = false;
  private String encoding ;

  static final public byte MINIMIZE_NONE = 0;
  static final public byte MINIMIZE_EMPTY_ELEMENTS = 1;
  static final public byte MINIMIZE_EMPTY_ELEMENTS_HTML = 2;


  public XhtmlOutputHandler() {
    lineSeparator = System.getProperty("line.separator");
  }

  public XhtmlOutputHandler(Writer writer) {
    this.writer = writer;
  }

  public DocumentHandler init(Destination dest, AttributeList atts) throws IOException {
    String mediaType = atts.getValue("media-type");
    if (mediaType == null)
      mediaType = "text/html";
    encoding = atts.getValue("encoding");
    if (encoding == null) {
      // not all Java implementations support ASCII
      writer = dest.getWriter(mediaType, "iso-8859-1");
      // use character references for non-ASCII characters
      maxRepresentableChar = '\u007F';
    }
    else {
      writer = dest.getWriter(mediaType, encoding);
      encoding = dest.getEncoding();
      if (encoding.equalsIgnoreCase("iso-8859-1"))
	maxRepresentableChar = '\u00FF';
      else if (encoding.equalsIgnoreCase("us-ascii"))
	maxRepresentableChar = '\u007F';
    }
    this.keepOpen = dest.keepOpen();
    if ("yes".equals(atts.getValue("omit-xml-declaration")))
      omitXmlDeclaration = true;
    this.standalone = atts.getValue("standalone");
    this.doctypeSystem = atts.getValue("doctype-system");
    this.doctypePublic = atts.getValue("doctype-public");
    if (this.doctypeSystem != null || this.doctypePublic != null)
      outputDoctype = true;
    if ("yes".equals(atts.getValue("indent")))
      return new Indenter(this, this);
    return this;
  }

  public void setMinimize(byte minimize) {
    this.minimize = minimize;
  }

  public void startDocument() throws SAXException {
    if (!omitXmlDeclaration) {
      write("<?xml version=\"1.0\" encoding=\"");
      write(encoding);
      write("\"");
      if (standalone != null) {
	write(" standalone=\"");
	write(standalone);
	write('"');
      }
      write("?>");
      write(lineSeparator);
    }
  }

  public void characters(char cbuf[], int off, int len) throws SAXException {
    if (len == 0)
      return;
    if (inStartTag)
      finishStartTag();
    do {
      char c = cbuf[off++];
      switch (c) {
      case '\n':
	write(lineSeparator);
	break;
      case '&':
	write("&amp;");
	break;
      case '<':
	write("&lt;");
	break;
      case  '>':
	write("&gt;");
	break;
      default:
	if (c < 0x80)
	  write(c);
	else {
	    write(c);
	}
      }
    } while (--len > 0);
  }

  public void rawCharacters(String chars) throws SAXException {
    if (inStartTag)
      finishStartTag();
    write(chars);
  }

  public void ignorableWhitespace (char ch[], int start, int length)
    throws SAXException {
    for (; length > 0; length--, start++)
      write(ch[start]);
  }

  public void startElement(String name, AttributeList atts) throws SAXException {
    if (inStartTag)
      finishStartTag();
    if (outputDoctype) {
      outputDoctype = false;
      write("<!DOCTYPE ");
      write(name);
      if (doctypePublic != null) {
	write(" PUBLIC ");
	char lit = doctypePublic.indexOf('"') >= 0 ?'\'' : '"';
	write(lit);
	write(doctypePublic);
	write(lit);
      }
      else
	write(" SYSTEM");
      if (doctypeSystem != null) {
	char lit = doctypePublic.indexOf('"') >= 0 ?'\'' : '"';
	write(' ');
	write(lit);
	write(doctypeSystem);
	write(lit);
      }
      write('>');
      write(lineSeparator);
    }
    write('<');
    write(name);
    int n = atts.getLength();
    for (int i = 0; i < n; i++) {
      write(' ');
      write(atts.getName(i));
      write('=');
      write('"');
      attributeValue(atts.getValue(i));
      write('"');
    }
    inStartTag = true;
  }

  protected void attributeValue(String value) throws SAXException {
    int valueLength = value.length();
    for (int j = 0; j < valueLength; j++) {
      char c = value.charAt(j);
      switch (c) {
      case '\n':
	write("&#10;");
	break;
      case '&':
	write("&amp;");
	break;
      case '<':
	write("&lt;");
	break;
      case '"':
	write("&quot;");
	break;
      case '\r':
	write("&#13;");
	break;
      case '\t':
	write("&#9;");
	break;
      default:
	if (c < 0x80)
	  write(c);
	else {
	    write(c);
	}
	break;
      }
    }
  }

  private final void finishStartTag() throws SAXException {
    inStartTag = false;
    write('>');
  }

  public void processingInstruction(String target, String data)
    throws SAXException {
    if (target == null) {
      comment(data);
      return;
    }
    if (inStartTag)
      finishStartTag();
    write('<');
    write('?');
    write(target);
    if (data.length() > 0) {
      write(' ');
      writeMarkup(data);
    }
    write('?');
    write('>');
  }

  public void markup(String chars) throws SAXException {
    if (inStartTag)
      finishStartTag();
    writeMarkup(chars);
  }

  public void comment(String body) throws SAXException {
    if (inStartTag)
      finishStartTag();
    write("<!--");
    writeMarkup(body);
    write("-->");
  }

  private void writeMarkup(String str) throws SAXException {
    int len = str.length();
    for (int i = 0; i < len; i++) {
      char c = str.charAt(i);
      if (c == '\n')
	write(lineSeparator);
      else if (c < 0x80)
	write(c);
      else {
	  write(c);
      }
    }
  }

  private final void flushBuf() throws SAXException {
    try {
      writer.write(buf, 0, bufUsed);
    }
    catch (IOException e) {
      throw new SAXException(e);
    }
    bufUsed = 0;
  }

  public void setDocumentLocator(Locator loc) { }

  public void endDocument() throws SAXException {
    write(lineSeparator);
    if (bufUsed != 0)
      flushBuf();
    try {
      if (keepOpen)
	writer.flush();
      else
	writer.close();
    }
    catch (IOException e) {
      throw new SAXException(e);
    }
    writer = null;
    buf = null;
  }

  private final void write(String s) throws SAXException {
    int start = 0;
    int len = s.length();
    int avail = buf.length - bufUsed;
    while (avail < len) {
      s.getChars(start, start + avail, buf, bufUsed);
      bufUsed = buf.length;
      flushBuf();
      start += avail;
      len -= avail;
      avail = buf.length;
    }
    s.getChars(start, start + len, buf, bufUsed);
    bufUsed += len;
  }

  private final void write(char b) throws SAXException {
    if (bufUsed == buf.length)
      flushBuf();
    buf[bufUsed++] = b;
  }


// Modified stuff to handle XHTML

  public void endElement(String name) throws SAXException {
  	int flags = getElementTypeFlags(name);
	if (inStartTag) {
		inStartTag = false;
		if ((flags & EMPTY_CONTENT) != 0) {
			write(' ');
			write('/');
			write('>');
			return;
		}
		write('>');
		write(' ');
	}
    write('<');
    write('/');
    write(name);
    write('>');
  }


}
